(* Msp.sml -- prelude for ML Server Pages
   sestoft@dina.kvl.dk 2000-11-06 version 0.8
 *)

(* Efficiently concatenable word sequences *)

datatype wseq = 
    Empty                               (* The empty sequence         *)
  | Nl                                  (* Newline                    *)
  | $ of string                         (* A string                   *)
  | $$ of string list                   (* A sequence of strings      *)
  | && of wseq * wseq;                  (* Concatenation of sequences *)

infix &&

fun prmap f []       = Empty 
  | prmap f (x1::xr) = 
    let fun loop y1 []       = f y1
	  | loop y1 (y2::yr) = f y1 && loop y2 yr 
    in loop x1 xr end

fun prsep sep f []       = Empty 
  | prsep sep f (x1::xr) = 
    let fun loop y1 []       = f y1
	  | loop y1 (y2::yr) = f y1 && sep && loop y2 yr 
    in loop x1 xr end

fun flatten Empty acc      = acc
  | flatten Nl    acc      = "\n" :: acc
  | flatten ($ s) acc      = s :: acc
  | flatten ($$ ss) acc    = List.@(ss, acc)
  | flatten (s1 && s2) acc = flatten s1 (flatten s2 acc)
val flatten = fn seq => String.concat(flatten seq []);
    
fun printseq Empty      = ()
  | printseq Nl         = TextIO.print "\n"
  | printseq ($ s)      = TextIO.print s
  | printseq ($$ ss)    = List.app TextIO.print ss
  | printseq (s1 && s2) = (printseq s1; printseq s2);
    
fun vec2list vec = Vector.foldr op:: [] vec

(* CGI parameter access shorthands *)

exception ParamMissing of string
exception NotInt of string * string

fun % fnm = 
    case Mosmlcgi.cgi_field_string fnm of 
	NONE   => raise ParamMissing "fnm"
      | SOME v => v

fun %? fnm = Option.isSome(Mosmlcgi.cgi_field_string fnm)

fun %# fnm =
    case Mosmlcgi.cgi_field_string fnm of 
	NONE   => raise ParamMissing fnm
      | SOME v => (case Int.fromString v of
		       NONE   => raise NotInt(fnm, v)
		     | SOME i => i)
		
fun %%(fnm, dflt)  = Option.getOpt(Mosmlcgi.cgi_field_string fnm, dflt)

fun %%#(fnm, dflt) = 
    Option.getOpt(Option.mapPartial Int.fromString
		  (Mosmlcgi.cgi_field_string fnm), dflt)

(* HTML generic marks *)

fun mark0 tag = $$["<", tag, ">"]
fun mark0a attr tag = $$["<", tag, " ", attr, ">"]
fun mark1 tag seq = $$["<", tag, ">"] && seq && $$["</", tag, ">"]
fun mark1a tag attr seq = 
    $$["<", tag, " ", attr, ">"] && seq && $$["</", tag, ">"]
fun comment seq = $"<!--" && seq && $"-->"

(* HTML documents and headers *)

fun html seq = $"<html>" && seq && $"</html>"
fun head seq = $"<head>" && seq && $"</head>"
fun title seq = $"<title>" && seq && $"</title>"
fun body seq = $"<body>" && seq && $"</body>"
fun bodya attr seq = $$["<body ", attr, ">"] && seq && $"</body>"
fun htmldoc tit bod = html (head (title tit) && body bod)

(* HTML headings and vertical format *)

fun h1 seq = $"<h1>" && seq && $"</h1>"
fun h2 seq = $"<h2>" && seq && $"</h2>"
fun h3 seq = $"<h3>" && seq && $"</h3>"
fun h4 seq = $"<h4>" && seq && $"</h4>"
fun h5 seq = $"<h5>" && seq && $"</h5>"
fun h6 seq = $"<h6>" && seq && $"</h6>"

fun p seq = $"<p>" && seq && $"</p>"
fun pa attr seq = $$["<P ", attr, ">"] && seq && $"</P>"
fun divi seq = $"<div>" && seq && $"</div>"
fun divia attr seq = $$["<div ", attr, ">"] && seq && $"</div>"
fun blockquote seq = $"<blockquote>" && seq && $"</blockquote>"
fun blockquotea attr seq = 
    $$["<blockquote ", attr, ">"] && seq && $"</blockquote>"
fun center seq = $"<center>" && seq && $"</center>"
fun address seq = $"<address>" && seq && $"</address>"
fun pre seq = $"<pre>" && seq && $"</pre>"

val br = $"<br>"
fun bra attr = $$["<br ", attr, ">"]
val hr = $"<hr>"
fun hra attr = $$["<hr ", attr, ">"]

(* HTML anchors and hyperlinks *)

fun ahref link seq = $$["<a href=\"", link, "\">"] && seq && $"</a>"
fun ahrefa link attr seq = 
    $$["<a href=\"", link, "\" ", attr, ">"] && seq && $"</a>"
fun aname name seq = $$["<a name=\"", name, "\">"] && seq && $"</a>"

(* HTML text formats and style *)

fun em seq = $"<em>" && seq && $"</em>"
fun strong seq = $"<strong>" && seq && $"</strong>"
fun code seq = $"<code>" && seq && $"</code>"
fun tt seq = $"<tt>" && seq && $"</tt>"
fun sub seq = $"<sub>" && seq && $"</sub>"
fun sup seq = $"<sup>" && seq && $"</sup>"
fun fonta attr seq = $$["<font ", attr, ">"] && seq && $"</font>"

(* HTML lists *)

fun ul seq = $"<ul>" && seq && $"</ul>"
fun ula attr seq = $$["<ul ", attr, ">"] && seq && $"</ul>"
fun ol seq = $"<ol>" && seq && $"</ol>"
fun ola attr seq = $$["<ol ", attr, ">"] && seq && $"</ol>"
fun li seq = $"<li>" && seq && $"</li>"

fun dl seq = $"<dl>" && seq && $"</dl>"
fun dla attr seq = $$["<dl ", attr, ">"] && seq && $"</dl>"
fun dt seq = $"<dt>" && seq && $"</dt>"
fun dd seq = $"<dd>" && seq && $"</dd>"

(* HTML tables *)

fun tr seq = $"<tr>" && seq && $"</tr>"
fun tra attr seq = $$["<tr ", attr, ">"] && seq && $"</tr>"
fun td seq = $"<td>" && seq && $"</td>"
fun tda attr seq = $$["<td ", attr, ">"] && seq && $"</td>"
fun th seq = $"<th>" && seq && $"</th>"
fun tha attr seq = $$["<th ", attr, ">"] && seq && $"</th>"
fun table seq = $"<table>" && seq && $"</table>"
fun tablea attr seq = $$["<table ", attr, ">"] && seq && $"</table>"
fun caption seq = $"<caption>" && seq && $"</caption>"
fun captiona attr seq = $$["<caption ", attr, ">"] && seq && $"</caption>"

(* HTML images and image maps *)

fun img src  = $$["<img src=\"", src, "\">"]
fun imga src attr = $$["<img src=\"", src, "\" ", attr, ">"]
fun map nam seq = $$["<map name=\"", nam, "\">"] && seq && $"</map>"
fun mapa nam attr seq = 
    $$["<map name=\"", nam, "\" ", attr, ">"] && seq && $"</map>"
fun area { shape, href, coords, alt } = 
    $$["<area shape=\"", shape, "\" coords=\"", coords, "\" "]
    && (case href of NONE => $"nohref" | SOME r => $$["href=\"", r, "\" "])
    && (case alt  of NONE => Empty | SOME a => $$["alt=\"", a, "\""])

(* HTML forms etc *)

fun form action seq = $$["<form action=\"", action, "\">"] && seq && $"</form>"
fun forma action attr seq = 
    $$["<form action=\"", action, "\" ", attr, ">"] && seq && $"</form>"
fun input typ = $$["<input type=", typ, ">"]
fun inputa typ attr = $$["<input type=", typ, " ", attr, ">"]
fun intext name attr = $$["<input type=text name=\"", name, "\" ", attr, ">"]
fun inpassword name attr = 
       $$["<input type=password name=\"", name, "\" ", attr, ">"]
fun incheckbox {name, value} attr = 
       $$["<input type=checkbox value=\"", value, "\" name=\"", name, 
	  "\" ", attr, ">"]
fun inradio {name, value} attr = 
       $$["<input type=radio value=\"", value, "\" name=\"", name, 
	  "\" ", attr, ">"]
fun inreset value attr = 
       $$["<input type=reset value=\"", value, "\" ", attr, ">"]
fun insubmit value attr = 
       $$["<input type=submit value=\"", value, "\" ", attr, ">"]
fun inhidden {name, value} = 
       $$["<input type=hidden name=\"", name, "\" value=\"", value, "\">"]
fun textarea name seq = 
	  $$["<textarea name=\"", name, "\">"] && seq && $"</textarea>"
fun textareaa name attr seq = 
	  $$["<textarea name=\"", name, "\" ", attr, ">"] 
	  && seq && $"</textarea>"
fun select name attr seq = 
	  $$["<select name=\"", name, "\" ", attr, ">"] && seq && $"</select>"
fun option value = $$["<option value=\"", value, "\">"]

(* HTML frames and framesets *)

fun frameset attr seq = $$["<frameset ", attr, ">"] && seq && $"</frameset>"
fun frame { src, name } = $$["<frame src=\"", src, "\" name=\"", name, "\">"] 
fun framea { src, name } attr = 
	  $$["<frame src=\"", src, "\" name=\"", name, "\" ", attr, ">"] 

(* HTML encoding *)

fun urlencode s : string =
    let fun encode #" " = "+"
	  | encode #"-" = "-"
	  | encode #"_" = "_"
	  | encode #"." = "."
	  | encode c    = 
	    if Char.isAlphaNum c then String.str c 
	    else "%" ^ StringCvt.padLeft #"0" 2 
                       (Int.fmt StringCvt.HEX (Char.ord c))
    in String.translate encode s end

(* Maybe this should create a wseq instead, to avoid creating a long 
   string by concatenation *)

fun htmlencode s : string =
    let fun encode #"<" = "&lt;"
	  | encode #">" = "&gt;"
	  | encode #"&" = "&amp;"
	  | encode c    = String.str c
    in String.translate encode s end
